#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteCode.h"
#include "KiteIdentifier.h"
#include "KiteTree.h"
#include "KiteList.h"
#include "KiteBoolean.h"

using namespace KiteVM;

KiteReference<KiteObject> KiteCode::classinfo;

KiteReference<KiteObject> KiteCode_eval(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	vm.syms.removeContext();
	KiteReference<KiteObject> ret(((KiteCode&)(*thisObj)).call(vm, param));
	vm.syms.addContext();
	return ret;
}

KiteReference<KiteObject> KiteCode_catch(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	vm.syms.removeContext();
	KiteReference<KiteObject> ret(((KiteCode&)(*thisObj)).call(vm, param));
	vm.syms.addContext();
	
	if (typeid(*ret) == typeid(KiteTree)) {
		// Run stuff in catch block.
		vm.syms.addVar("exception", ret);
        if (typeid(*param) == typeid(KiteCode)) {
            return ((KiteCode&)(*param)).call(vm, KiteReference<KiteObject>(new KiteList(false)));
        } else {
            return param;
        }
	}

	return ret;
}

KiteReference<KiteObject> KiteCode_while(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	vm.syms.removeContext();
	KiteReference<KiteObject> ret, cond;
	
    if (typeid(*param) != typeid(KiteCode)) {
        // Need a code object for the conditional.
        return KiteReference<KiteObject>(new KiteTree("Code block required as parameter for while"));
    }
    
begin:
	ret = ((KiteCode&)(*thisObj)).call(vm, param);
	if (typeid(*ret) == typeid(KiteTree)) {
		// Exception; return now
		vm.syms.addContext();
		return ret;
	}

	// Loop conditional
	cond = ((KiteCode&)(*param)).call(vm, new KiteList(false));
	if (typeid(*cond) != typeid(KiteBoolean)) {
		// Condition must return boolean
		vm.syms.addContext();
		return KiteReference<KiteObject>(new KiteTree("While condition must return boolean"));
	}

	// Continue if condition is true
	if (((KiteBoolean&)(*cond)).boolVal()) goto begin;

	// Else, return what the final iteration returned.
	vm.syms.addContext();
	return ret;
}

KiteReference<KiteObject> KiteCode_for(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteReference<KiteObject> ret(new KiteBoolean(false)), newlist, item;
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort >::iterator i;
	
	// Convert parameter into a singleton list, if needed
	if (typeid(*param) != typeid(KiteList)) {
		newlist = KiteReference<KiteObject>(new KiteList(param));
	} else {
		newlist = param;
	}

	// Get items from list and iterate through them
    for (int j = 0; j < ((KiteList&)(*newlist)).getArray().size(); j++) {
        vm.syms.addContext();
		vm.syms.addVar("key", KiteReference<KiteObject>(new KiteNumber(j)));
		vm.syms.addVar("value", ((KiteList&)(*newlist)).getArray().at(j));

		// Call function on iterated stuff
		ret = ((KiteCode&)(*thisObj)).call(vm, new KiteList(false));
		if (typeid(*ret) == typeid(KiteTree)) {
			// Exception; return now
			vm.syms.removeContext();
			return ret;
		}
		vm.syms.removeContext();
    }
    
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort > &items = ((KiteList&)(*newlist)).getItems();
	for (i = items.begin(); i != items.end(); i++) {
		vm.syms.addContext();
		vm.syms.addVar("key", (*i).first);
		vm.syms.addVar("value", (*i).second);

		// Call function on iterated stuff
		ret = ((KiteCode&)(*thisObj)).call(vm, new KiteList(false));
		if (typeid(*ret) == typeid(KiteTree)) {
			// Exception; return now
			vm.syms.removeContext();
			return ret;
		}
		vm.syms.removeContext();
	}

	// Return what the final iteration returned.
	return ret;
}

void KiteCode::initialize() {
	syms["eval"] = KiteReference<KiteObject>(new KiteCode(KiteCode_eval));
	syms["catch"] = KiteReference<KiteObject>(new KiteCode(KiteCode_catch));
	syms["while"] = KiteReference<KiteObject>(new KiteCode(KiteCode_while));
	syms["for"] = KiteReference<KiteObject>(new KiteCode(KiteCode_for));
}

KiteReference<KiteObject> KiteCode::call(KiteVM::VM &vm, KiteReference<KiteObject> param, bool overwriteArgs) {
	if (funcPtr) {
		return (*funcPtr)(vm, param);
	} else {
		if (overwriteArgs) vm.syms.addVar("args", param);

		// Save # of items on stack so we can return the stack to how it was
		int items = vm.stack.size();

		KiteReference<KiteObject> ret = vm.run(instr);

		// Return stack to how it was before
		for (int i = vm.stack.size(); i > items; i--) {
			vm.stack.pop();
		}

		return ret;
	}
}

KiteCode::~KiteCode() {
	// nothing needed
}

//	virtual KiteReference<KiteObject> dereference(VM &vm, KiteIdentifier &id);
