/*
 *  KiteDyamicLoader.cpp
 *  kite
 *
 *  Created by Mooneer Salem on 1/5/07.
 *  Copyright 2006 __MyCompanyName__. All rights reserved.
 *
 */

#include "KiteDynamicLoader.h"
#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteReference.h"
#include "KiteTree.h"

KiteDynamicLoader::KiteDynamicLoader(const char *filename) {
	// Load library
	string actualname = filename;
#ifdef WIN32
	actualname += ".dll";
	WCHAR *wszName = new WCHAR[actualname.size() + 1];
	MultiByteToWideChar(CP_ACP, 0, actualname.c_str(), actualname.size(), wszName, sizeof(wszName)/sizeof(wszName[0]));
	library = LoadLibrary(wszName);
	delete[] wszName;
#else
	actualname += ".so";
	library = dlopen(actualname.c_str(), RTLD_LAZY);
#endif
	if (library == NULL) {
		char buf[256];
#ifdef WIN32
		sprintf(buf, "%d", GetLastError());
#else
		sprintf(buf, "%s", dlerror());
#endif
		throw KiteReference<KiteObject>(new KiteTree(string("Could not load ") + string(filename) + string(": error ") + string(buf)));
	}
}

KiteDynamicLoader::~KiteDynamicLoader() {
	// Unload library
#ifdef WIN32
	FreeLibrary(library);
#else
	dlclose(library);
#endif
}

void *KiteDynamicLoader::findFunction(const char *name) {
	// Find function in current library
	void *ret;
#ifdef WIN32
	ret = GetProcAddress(library, name);
#else
	ret = dlsym(library, name);
#endif // WIN32

	if (!ret) {
		char buf[256];
#ifdef WIN32
		sprintf(buf, "%d", GetLastError());
#else
		sprintf(buf, "%s", dlerror());
#endif // WIN32
		throw KiteReference<KiteObject>(new KiteTree(string("Could not load ") + string(name) + string(": error ") + string(buf)));
	}

	return ret;
}

