/*
 *  KiteExternalFunction.cpp
 *  kite
 *
 *  Created by Mooneer Salem on 12/28/06.
 *  Copyright 2006 __MyCompanyName__. All rights reserved.
 *
 */

#include <errno.h>
#include <dlfcn.h>
#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteCode.h"
#include "KiteString.h"
#include "KiteTree.h"
#include "KiteIdentifier.h"
#include "KiteList.h"
#include "KiteExternalFunction.h"
#include "KiteBoolean.h"

using namespace std;

// Maximum number of parameters
#define EXTERN_MAX_PARAMETERS 3
typedef long double maximum_sized_param;
struct parameters {
    unsigned char param[sizeof(maximum_sized_param) * EXTERN_MAX_PARAMETERS];
};
typedef void (*voidFunction)(struct parameters parm);
typedef short (*shortFunction)(struct parameters parm);
typedef int (*intFunction)(struct parameters parm);
typedef double (*doubleFunction)(struct parameters parm);
typedef float (*floatFunction)(struct parameters parm);
typedef char *(*stringFunction)(struct parameters parm);
typedef char (*charFunction)(struct parameters parm);
typedef long (*longFunction)(struct parameters parm);
typedef long double (*longDoubleFunction)(struct parameters parm);

KiteReference<KiteObject> KiteExternalFunction::classinfo;
static string strings[EXTERN_MAX_PARAMETERS];

int parameterToBuffer(int argNumber, KiteVM::VM &vm, char type, char *current, KiteReference<KiteObject> param) {
    int size;
    char cval;
    short sval;
    int ival;
    long lval;
    float fval;
    double dval;
    long double ldval;
    char *ptr;

    switch(type) {
        case 's':
            size = sizeof(char*);
            strings[argNumber] = ((KiteString&)(*coerceString(vm, param))).getString();
            ptr = const_cast<char*>(strings[argNumber].c_str());
            memcpy(current, &ptr, size);
            break;
        case 'c':
            size = sizeof(char);
            cval = (char)((KiteNumber&)(*coerceNumber(vm, param))).intValue();
            memcpy(current, &cval, size);
            break;
        case 'h':
            size = sizeof(short);
            sval = (short)((KiteNumber&)(*coerceNumber(vm, param))).intValue();
            memcpy(current, &sval, size);
            break;
        case 'i':
            size = sizeof(int);
            ival = (int)((KiteNumber&)(*coerceNumber(vm, param))).intValue();
            memcpy(current, &ival, size);
            break;
        case 'l':
            size = sizeof(long);
            lval = (long)((KiteNumber&)(*coerceNumber(vm, param))).intValue();
            memcpy(current, &lval, size);
            break;
        case 'f':
            size = sizeof(float);
            fval = (float)((KiteNumber&)(*coerceNumber(vm, param))).floatValue();
            memcpy(current, &fval, size);
            break;
        case 'd':
            size = sizeof(double);
            dval = (double)((KiteNumber&)(*coerceNumber(vm, param))).floatValue();
            memcpy(current, &dval, size);
            break;
        case 'o':
            size = sizeof(long double);
            ldval = (long double)((KiteNumber&)(*coerceNumber(vm, param))).floatValue();
            memcpy(current, &ldval, size);
            break;
    }

    return size;
}

KiteReference<KiteObject> KiteExternalFunction_call(KiteVM::VM &vm, KiteReference<KiteObject> param) {
    KiteExternalFunction &obj = (KiteExternalFunction&)(*THIS_OBJECT());
    struct parameters passed;
    char *current = (char*)&passed;
    char ret = obj.getReturnType()[0];
    char *str, x;
    short sx;
    int ix;
    long lx;
    float fx;
    double dx;
    long double ldx;

    if ((typeid(*param) != typeid(KiteList) && obj.getArgumentTypes().length() > 1) ||
        ((KiteList&)(*param)).getArray().size() < obj.getArgumentTypes().length()) {
        return KiteReference<KiteObject>(new KiteTree("Not enough arguments"));
    }

    // Convert and pass each parameter one at a time.
    if (typeid(*param) != typeid(KiteList)) {
        current += parameterToBuffer(0, vm, obj.getArgumentTypes()[0], current, param);
    } else {
        KiteList &list = (KiteList&)(*param);
        for (int i = 0; i < obj.getArgumentTypes().length(); i++) {
            KiteReference<KiteObject> val = list.getArray()[i];
            current += parameterToBuffer(i, vm, obj.getArgumentTypes()[i], current, val);
        }
    }

    // Call function and return.
    switch(ret) {
        case 'v':
            (*(voidFunction)(obj.getFuncPointer()))(passed);
            return KiteReference<KiteObject>(new KiteBoolean(false));
        case 's':
            str = (*(stringFunction)(obj.getFuncPointer()))(passed);
            return KiteReference<KiteObject>(new KiteString(str));
        case 'c':
            x = (*(charFunction)(obj.getFuncPointer()))(passed);
            return KiteReference<KiteObject>(new KiteNumber(x));
        case 'h':
            sx = (*(shortFunction)(obj.getFuncPointer()))(passed);
            return KiteReference<KiteObject>(new KiteNumber(sx));
        case 'i':
            ix = (*(intFunction)(obj.getFuncPointer()))(passed);
            return KiteReference<KiteObject>(new KiteNumber(ix));
        case 'l':
            lx = (*(longFunction)(obj.getFuncPointer()))(passed);
            return KiteReference<KiteObject>(new KiteNumber((int)lx));
        case 'f':
            fx = (*(floatFunction)(obj.getFuncPointer()))(passed);
            return KiteReference<KiteObject>(new KiteNumber(fx));
        case 'd':
            dx = (*(doubleFunction)(obj.getFuncPointer()))(passed);
            return KiteReference<KiteObject>(new KiteNumber(dx));
        case 'o':
            ldx = (*(longDoubleFunction)(obj.getFuncPointer()))(passed);
            return KiteReference<KiteObject>(new KiteNumber((double)ldx));
    }

    return KiteReference<KiteObject>(new KiteBoolean(false));
}

KiteReference<KiteObject>
KiteExternalFunction::create(KiteVM::VM &vm, KiteReference<KiteObject> base, KiteReference<KiteObject> param) {
    // We need four arguments: the name of the library file, the return type,
    // the argument types and the function name.
    KiteReference<KiteObject> libName, retType, argTypes, funcName;
    
    if (typeid(*param) != typeid(KiteList) || ((KiteList&)(*param)).getArray().size() < 4) {
        return KiteReference<KiteObject>(new KiteTree("Not enough arguments"));
    }
    
    KiteList &list = (KiteList&)(*param);
    libName = coerceString(vm, list.getArray()[0]);
    retType = coerceString(vm, list.getArray()[1]);
    argTypes = coerceString(vm, list.getArray()[2]);
    funcName = coerceString(vm, list.getArray()[3]);
    
    KiteExternalFunction *s = new KiteExternalFunction(((KiteString&)(*libName)).getString(), ((KiteString&)(*retType)).getString(),
                                   ((KiteString&)(*argTypes)).getString(), ((KiteString&)(*funcName)).getString());
    
    if (s->libraryFile == NULL || s->funcPtr == NULL) {
        // Throw an exception, error found
        KiteTree *t = new KiteTree(string("[loader] ") + dlerror());
        delete s;
        return KiteReference<KiteObject>(t);
    }
    
    // Otherwise, return the object we just created
    return KiteReference<KiteObject>(s);
}

void KiteExternalFunction::initialize() {
	syms["call"] = KiteReference<KiteObject>(new KiteCode(KiteExternalFunction_call));
}
