/*
 *  KiteExternalFunction.h
 *  kite
 *
 *  Created by Mooneer Salem on 3/31/07.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#include <dlfcn.h>

class KiteExternalFunction : public KiteObject {
public:
	KiteExternalFunction()
		: KiteObject(false) {
        inherits = KiteObject::getClassinfo();
        this->type = "loader";
		initialize();
	}

    KiteExternalFunction(string libName, string retType, string argTypes, string funcName)
        : KiteObject(false), instance(true), libraryFile(NULL), funcPtr(NULL) {
        inherits = KiteExternalFunction::getClassinfo();
        this->type = "loader";
        
        // Load file using dlopen. If zero-length, load from libc
        if (libName.length() == 0) {
            libraryFile = dlopen(NULL, RTLD_NOW);
        } else {
            libraryFile = dlopen(libName.c_str(), RTLD_NOW);
        }
        if (!libraryFile) return;

        // Load function using given library file.
        funcPtr = dlsym(libraryFile, funcName.c_str());

        // Initialize everything else
        this->retType = retType;
        this->argTypes = argTypes;
    }

    virtual ~KiteExternalFunction() {
        // Unload library
        if (instance && libraryFile) {
            dlclose(libraryFile);
        }
    }
    
    virtual KiteReference<KiteObject> getClassinfo() {
        if (classinfo.getObj() == NULL) {
            classinfo = new KiteExternalFunction();
        }
        return classinfo;
    }
    
    virtual KiteReference<KiteObject>
	create(KiteVM::VM &vm, KiteReference<KiteObject> base, KiteReference<KiteObject> param);
    
    inline string &getReturnType() { return retType; }
    inline string &getArgumentTypes() { return argTypes; }
    inline void *getFuncPointer() { return funcPtr; }
private:
	bool instance;
	void initialize();
    static KiteReference<KiteObject> classinfo;
    void *libraryFile, *funcPtr;
    string retType, argTypes;
};
