#include "KiteVM.h"
#include "KiteNumber.h"
#include "KiteString.h"
#include "KiteReference.h"
#include "KiteIdentifier.h"
#include "KiteCode.h"
#include "KiteBoolean.h"
#include "KiteTree.h"
#include "KiteFile.h"
#include "KiteList.h"

using namespace KiteVM;

KiteReference<KiteObject> KiteFile_write(VM &vm, KiteReference<KiteObject> param) {
	KiteFile &fp = ((KiteFile&)(*THIS_OBJECT()));
	KiteReference<KiteObject> str = coerceString(vm, param);

	fp.getOutput() << ((KiteString&)(*str)).getString();
	return THIS_OBJECT();
}

KiteReference<KiteObject> KiteFile_read(VM &vm, KiteReference<KiteObject> param) {
	KiteFile &fp = ((KiteFile&)(*THIS_OBJECT()));
	KiteNumber &num = (KiteNumber&)(*coerceNumber(vm, param));
	int i = num.intValue();
	char *str = new char[i];

	fp.getInput().read(str, i);
	KiteReference<KiteObject> ret(new KiteString(str));
	delete[] str;
	return ret;
}

KiteReference<KiteObject> KiteFile_readline(VM &vm, KiteReference<KiteObject> param) {
	KiteFile &fp = ((KiteFile&)(*THIS_OBJECT()));
	char buf[1024];

	fp.getInput().getline(buf, 1024);
	KiteReference<KiteObject> ret(new KiteString(buf));
	return ret;
}

KiteReference<KiteObject> KiteFile_close(VM &vm, KiteReference<KiteObject> param) {
	KiteFile &fp = ((KiteFile&)(*THIS_OBJECT()));

	if (dynamic_cast<fstream&>(fp.getInput()).is_open()) {
		dynamic_cast<fstream&>(fp.getInput()).close();
	}
	if (dynamic_cast<fstream&>(fp.getOutput()).is_open()) {
		dynamic_cast<fstream&>(fp.getOutput()).close();
	}

	return KiteReference<KiteObject>(new KiteBoolean(true));
}

KiteReference<KiteObject> KiteFile_seek(VM &vm, KiteReference<KiteObject> param) {
	KiteFile &fp = ((KiteFile&)(*THIS_OBJECT()));
    KiteReference<KiteObject> p(coerceNumber(vm, param));
    KiteNumber &n = (KiteNumber&)(*p);
    
	if (dynamic_cast<fstream&>(fp.getInput()).is_open()) {
        dynamic_cast<fstream&>(fp.getInput()).seekg(n.intValue(), ios_base::beg);
        return THIS_OBJECT();
    }

    return KiteReference<KiteObject>(new KiteTree("File must be open to seek"));
}

KiteReference<KiteObject> KiteFile_tell(VM &vm, KiteReference<KiteObject> param) {
	KiteFile &fp = ((KiteFile&)(*THIS_OBJECT()));
    
	if (dynamic_cast<fstream&>(fp.getInput()).is_open()) {
        return KiteReference<KiteObject>(new KiteNumber((int)dynamic_cast<fstream&>(fp.getInput()).tellg()));
    }

    return KiteReference<KiteObject>(new KiteTree("File must be open to get position"));
}

KiteReference<KiteObject> KiteFile_filesize(VM &vm, KiteReference<KiteObject> param) {
	KiteFile &fp = ((KiteFile&)(*THIS_OBJECT()));
    
	if (dynamic_cast<fstream&>(fp.getInput()).is_open()) {
        int origpos = dynamic_cast<fstream&>(fp.getInput()).tellg();
        dynamic_cast<fstream&>(fp.getInput()).seekg (0, ios::end);
        int length = dynamic_cast<fstream&>(fp.getInput()).tellg();
        dynamic_cast<fstream&>(fp.getInput()).seekg (origpos, ios::beg);

        return KiteReference<KiteObject>(new KiteNumber(length));
    }

    return KiteReference<KiteObject>(new KiteTree("File must be open to get size"));
}

void KiteFile::initialize() {
	syms["write"] = KiteReference<KiteObject>(new KiteCode(KiteFile_write));
	syms["read"] = KiteReference<KiteObject>(new KiteCode(KiteFile_read));
	syms["readline"] = KiteReference<KiteObject>(new KiteCode(KiteFile_readline));
    syms["seek"] = KiteReference<KiteObject>(new KiteCode(KiteFile_seek));
    syms["tell"] = KiteReference<KiteObject>(new KiteCode(KiteFile_tell));
    syms["filesize"] = KiteReference<KiteObject>(new KiteCode(KiteFile_filesize));
	syms["close"] = KiteReference<KiteObject>(new KiteCode(KiteFile_close));
}
