/*
 *  KiteMath.cpp
 *  kite
 *
 *  Created by Mooneer Salem on 12/27/06.
 *  Copyright 2006 __MyCompanyName__. All rights reserved.
 *
 */

#include <math.h>
#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteCode.h"
#include "KiteString.h"
#include "KiteTree.h"
#include "KiteIdentifier.h"
#include "KiteList.h"
#include "KiteMath.h"

using namespace std;

KiteReference<KiteObject> KiteMath::classinfo;

#define MATH_WRAPPER(name) \
    KiteReference<KiteObject> KiteMath_ ## name (VM &vm, KiteReference<KiteObject> param) { \
        KiteReference<KiteObject> number(coerceNumber(vm, param)); \
        KiteNumber &n = (KiteNumber&)(*number); \
        return KiteReference<KiteObject>(new KiteNumber(name(n.floatValue()))); \
    }

#define MATH_WRAPPER_INT(name) \
    KiteReference<KiteObject> KiteMath_ ## name (VM &vm, KiteReference<KiteObject> param) { \
        KiteReference<KiteObject> number(coerceNumber(vm, param)); \
        KiteNumber &n = (KiteNumber&)(*number); \
        return KiteReference<KiteObject>(new KiteNumber(name(n.intValue()))); \
    }

#define MATH_WRAPPER_TWOARG(name) \
    KiteReference<KiteObject> KiteMath_ ## name (VM &vm, KiteReference<KiteObject> param) { \
        if (typeid(*param) != typeid(KiteList) && ((KiteList&)(*param)).getArray().size() < 2) { \
            return KiteReference<KiteObject>(new KiteTree("Not enough arguments")); \
        } \
        KiteList &l = ((KiteList&)(*param)); \
        KiteReference<KiteObject> a1(coerceNumber(vm, l.getArray()[0])); \
        KiteNumber &n1 = ((KiteNumber&)(*a1)); \
        KiteReference<KiteObject> a2(coerceNumber(vm, l.getArray()[1])); \
        KiteNumber &n2 = ((KiteNumber&)(*a2)); \
        return KiteReference<KiteObject>(new KiteNumber(name(n1.floatValue(), n2.floatValue()))); \
    }
    
MATH_WRAPPER_INT(abs)
MATH_WRAPPER(acos)
MATH_WRAPPER(asin)
MATH_WRAPPER(atan)
MATH_WRAPPER(ceil)
MATH_WRAPPER(cos)
MATH_WRAPPER(cosh)
MATH_WRAPPER(exp)
MATH_WRAPPER(fabs)
MATH_WRAPPER_TWOARG(fmod)
MATH_WRAPPER(floor)
MATH_WRAPPER_TWOARG(ldexp)
MATH_WRAPPER(log)
MATH_WRAPPER(log10)
MATH_WRAPPER_TWOARG(pow)
MATH_WRAPPER(sin)
MATH_WRAPPER(sinh)
MATH_WRAPPER(sqrt)
MATH_WRAPPER(tan)
MATH_WRAPPER(tanh)

KiteReference<KiteObject> KiteMath_modf (VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> number(coerceNumber(vm, param));
    KiteNumber &n = (KiteNumber&)(*number);
    double integer, frac;
    KiteList *ret = new KiteList(false);
    
    frac = modf(n.floatValue(), &integer);
    ret->getArray().push_back(KiteReference<KiteObject>(new KiteNumber(integer)));
    ret->getArray().push_back(KiteReference<KiteObject>(new KiteNumber(frac)));
    
    return KiteReference<KiteObject>(ret);
}

KiteReference<KiteObject> KiteMath_frexp (VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> number(coerceNumber(vm, param));
    KiteNumber &n = (KiteNumber&)(*number);
    double mantissa, exponent;
    KiteList *ret = new KiteList(false);
    
    mantissa = modf(n.floatValue(), &exponent);
    ret->getArray().push_back(KiteReference<KiteObject>(new KiteNumber(mantissa)));
    ret->getArray().push_back(KiteReference<KiteObject>(new KiteNumber(exponent)));
    
    return KiteReference<KiteObject>(ret);
}

void KiteMath::initialize() {
	syms["abs"] = KiteReference<KiteObject>(new KiteCode(KiteMath_abs));
	syms["acos"] = KiteReference<KiteObject>(new KiteCode(KiteMath_acos));
	syms["asin"] = KiteReference<KiteObject>(new KiteCode(KiteMath_asin));
	syms["atan"] = KiteReference<KiteObject>(new KiteCode(KiteMath_atan));
	syms["ceil"] = KiteReference<KiteObject>(new KiteCode(KiteMath_ceil));
	syms["cos"] = KiteReference<KiteObject>(new KiteCode(KiteMath_cos));
	syms["cosh"] = KiteReference<KiteObject>(new KiteCode(KiteMath_cosh));
	syms["exp"] = KiteReference<KiteObject>(new KiteCode(KiteMath_exp));
	syms["fabs"] = KiteReference<KiteObject>(new KiteCode(KiteMath_fabs));
	syms["floor"] = KiteReference<KiteObject>(new KiteCode(KiteMath_floor));
	syms["fmod"] = KiteReference<KiteObject>(new KiteCode(KiteMath_fmod));
	syms["frexp"] = KiteReference<KiteObject>(new KiteCode(KiteMath_frexp));
	syms["ldexp"] = KiteReference<KiteObject>(new KiteCode(KiteMath_ldexp));
	syms["log"] = KiteReference<KiteObject>(new KiteCode(KiteMath_log));
	syms["log10"] = KiteReference<KiteObject>(new KiteCode(KiteMath_log10));
	syms["modf"] = KiteReference<KiteObject>(new KiteCode(KiteMath_modf));
	syms["pow"] = KiteReference<KiteObject>(new KiteCode(KiteMath_pow));
	syms["sin"] = KiteReference<KiteObject>(new KiteCode(KiteMath_sin));
	syms["sinh"] = KiteReference<KiteObject>(new KiteCode(KiteMath_sinh));
	syms["sqrt"] = KiteReference<KiteObject>(new KiteCode(KiteMath_sqrt));
	syms["tan"] = KiteReference<KiteObject>(new KiteCode(KiteMath_tan));
	syms["tanh"] = KiteReference<KiteObject>(new KiteCode(KiteMath_tanh));
	syms["pi"] = KiteReference<KiteObject>(new KiteNumber(3.14159265358979));
	syms["e"] = KiteReference<KiteObject>(new KiteNumber(exp(1.0)));
}


