#ifndef KITEOBJECT_H
#define KITEOBJECT_H

#include <map>
#include "KiteReference.h"

using namespace std;

class KiteIdentifier;
class KiteList;

extern bool compareIdentifiers(const KiteReference<KiteIdentifier> &r1, const KiteReference<KiteIdentifier> &r2);

struct KiteIdentifierCompare {
    bool operator()(const KiteReference<KiteIdentifier> &r1, const KiteReference<KiteIdentifier> &r2) const {
		return compareIdentifiers(r1, r2);
    }
};

class KiteObject {
public:
	KiteObject(); // default, "class info" constructor
    KiteObject(bool init);
	KiteObject(KiteVM::VM &vm, KiteReference<KiteObject> param); // class|new
	virtual ~KiteObject();

	inline bool isInstance() { return instance; }

	// Variable lookup/execute
	virtual KiteReference<KiteObject> call(KiteVM::VM &vm, KiteReference<KiteObject> param, bool overwriteArgs = true);
	inline virtual KiteReference<KiteObject> dereference(KiteVM::VM &vm, string id) {
        // Go through class hierachy, looking for object
        KiteObject *cur = this;

        while(cur) {
            if (cur->syms.count(id) > 0) {
                return cur->syms[id];
            }
            cur = &(*cur->inherits);
        }

        // We can't find the object, so time to create it.
        return syms[id];
    }

	inline virtual bool operator<(const KiteObject &rhs) const {
		return this < &rhs;
	}

	inline virtual bool operator==(const KiteObject &rhs) const {
		return this == &rhs;
	}

	inline virtual KiteReference<KiteObject> getClassinfo() {
        if (classinfo.getObj() == NULL) {
            classinfo = new KiteObject();
        }

		return classinfo;
	}

	virtual KiteReference<KiteObject>
	create(KiteVM::VM &vm, KiteReference<KiteObject> base, KiteReference<KiteObject> param);
	
	map<string, KiteReference<KiteObject> > &getSyms() { return syms; }
	inline virtual KiteReference<KiteObject> getInherits() { return inherits; }

    string type;
protected:
	KiteReference<KiteObject> inherits; // object inherited from
	static KiteReference<KiteObject> classinfo;
	bool instance;
	static bool initialized;

	map<string, KiteReference<KiteObject> > syms;

	void initialize();
};

#endif // KITEOBJECT_H
