#ifndef KITEREFERENCE_H
#define KITEREFERENCE_H

//#include <map>

class KiteIdentifier;

//using namespace std;

template<class T>
class KiteReference {
public:
    inline KiteReference() : refs(NULL), object(NULL) { }
    
    inline KiteReference(T *val) {
        if (val) {
            refs = new int(1);
            object = new T*(val);
        } else {
            object = NULL;
            refs = NULL;
        }
    }
    
    inline KiteReference(const KiteReference<T> &obj)
    : refs(obj.refs),
    object(obj.object) { 
        if (refs) (*refs)++;
    }
    
    inline virtual ~KiteReference() {
        destroy();
    };
    
    inline T *operator->() const {
        return (object ? *object : NULL);
    };
    
    inline T &operator*() const {
        return **object;
    };
    
    inline KiteReference<T> &operator=(const KiteReference<T> &in) {
        replace(in);
        return *this;
    }
    
    inline void replace(const KiteReference<T> &rhs) {
        destroy();
        object = rhs.object;
        refs = rhs.refs;
        if (refs) (*refs)++;
    }
    
    inline KiteReference<KiteIdentifier> referenceIdentifier() {
        KiteReference<KiteIdentifier> ret(NULL);
        ret.refs = refs;
        ret.object = (KiteIdentifier**)object;
        if (ret.refs) (*ret.refs)++;
        return ret;
    }
    
    inline T *operator=(const T *in) {
        if (object) {
            delete *object;
        } else {
            object = new T*();
            refs = new int(1);
        }
        *object = const_cast<T*>(in);
        return *object;
    }
    
    inline bool operator==(const KiteReference<T> &ptr) const {
        return (object ? **object : NULL) == (ptr.object ? **ptr.object : NULL);
    }
    
    inline bool operator==(const T *ptr) const {
        return (object ? *object : NULL) == ptr;
    }
    
    inline bool operator<(const KiteReference<T> &ptr) const {
        return (object ? **object : NULL) < (ptr.object ? **ptr.object : NULL);
    }
    
    inline int *getRefs() const { return refs; }
    inline T *getObj() const { return (object ? *object : NULL); }
    
public:
    int *refs;
    T **object;
    
private:
    inline void destroy() {
        if (refs) {
            (*refs)--;
            if (*refs == 0) {
                delete refs;
                delete *object;
                delete object;
            }
        }
    }
};

#endif // KITEREFERENCE_H
