/*
 *  KiteSocket.cpp
 *  kite
 *
 *  Created by Mooneer Salem on 12/28/06.
 *  Copyright 2006 __MyCompanyName__. All rights reserved.
 *
 */

#include <errno.h>
#ifdef WIN32
#include <winsock2.h>
#else
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif
#include <sys/types.h>
#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteCode.h"
#include "KiteString.h"
#include "KiteTree.h"
#include "KiteIdentifier.h"
#include "KiteList.h"
#include "KiteSocket.h"
#include "KiteBoolean.h"

using namespace std;

KiteReference<KiteObject> KiteSocket::classinfo;

#ifdef WIN32
bool KiteSocket::winsockInit = false;
#endif

KiteReference<KiteObject> KiteSocket_connect (VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> host, port;
    KiteSocket &s = (KiteSocket&)(*THIS_OBJECT());
    
    if (typeid(*param) != typeid(KiteList) || ((KiteList&)(*param)).getArray().size() < 2) {
        return KiteReference<KiteObject>(new KiteTree("Not enough arguments"));
    }
    
    KiteList &list = (KiteList&)(*param);
    host = coerceString(vm, list.getArray()[0]);
    port = coerceNumber(vm, list.getArray()[1]);
    
    string h = ((KiteString&)(*host)).getString();
    int p = ((KiteNumber&)(*port)).intValue();
    
    struct hostent *server;
    struct sockaddr_in serv_addr;
    server = gethostbyname(h.c_str());
    if (server == NULL) {
        goto error;
    }
    
#ifdef WIN32
#define bzero(a, b) memset(a, 0, b)
#define bcopy(a, b, c) memcpy(a, b, c)
#endif

    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    bcopy((char *)server->h_addr, 
         (char *)&serv_addr.sin_addr.s_addr,
         server->h_length);
    serv_addr.sin_port = htons(p);
    
    if (connect(s.getSocket(), (sockaddr*)&serv_addr, sizeof(serv_addr)) < 0) {
        goto error;
    }
    
    return KiteReference<KiteObject>(new KiteBoolean(true));
error:
    return KiteReference<KiteObject>(new KiteTree(string("[socket] ") + strerror(errno)));
}

KiteReference<KiteObject> KiteSocket_read (VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> size(coerceNumber(vm, param));
    KiteSocket &s = (KiteSocket&)(*THIS_OBJECT());
    KiteNumber &n = ((KiteNumber&)(*size));
    KiteReference<KiteObject> r;

    char *buf = new char[n.intValue() + 1];
    memset(buf, 0, n.intValue() + 1);
    
    int ret = recv(s.getSocket(), buf, n.intValue(), 0);
    if (ret == -1) goto error;

	r = new KiteString(buf);
	delete[] buf;
	return r;
error:
	delete[] buf;
    return KiteReference<KiteObject>(new KiteTree(string("[socket] ") + strerror(errno)));
}

KiteReference<KiteObject> KiteSocket_readline (VM &vm, KiteReference<KiteObject> param) {
    KiteSocket &s = (KiteSocket&)(*THIS_OBJECT());
    
    char buf[2];
    string ret;
    memset(buf, 0, 2);
    
    do {
        int n = recv(s.getSocket(), buf, 1, 0);
        if (n == -1) goto error;
        if (n == 0) break;
        
        buf[1] = '\0';
        ret += buf;
    } while(buf[0] != '\n');
     
    return KiteReference<KiteObject>(new KiteString(ret));
error:
    return KiteReference<KiteObject>(new KiteTree(string("[socket] ") + strerror(errno)));
}

KiteReference<KiteObject> KiteSocket_write (VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> data(coerceString(vm, param));
    KiteSocket &s = (KiteSocket&)(*THIS_OBJECT());
    KiteString &d = (KiteString&)(*data);
    
    int ret = send(s.getSocket(), d.getString().c_str(), d.getString().size(), 0);
    if (ret == -1) goto error;
    
    return KiteReference<KiteObject>(new KiteNumber(ret));
error:
    return KiteReference<KiteObject>(new KiteTree(string("[socket] ") + strerror(errno)));
}

KiteReference<KiteObject> KiteSocket_close (VM &vm, KiteReference<KiteObject> param) {
    KiteSocket &s = (KiteSocket&)(*THIS_OBJECT());
    
#ifdef WIN32
	shutdown(s.getSocket(), SD_BOTH);
#else
    shutdown(s.getSocket(), SHUT_RDWR);
#endif
    return KiteReference<KiteObject>(new KiteBoolean(true));
}

KiteReference<KiteObject>
KiteSocket::create(KiteVM::VM &vm, KiteReference<KiteObject> base, KiteReference<KiteObject> param) {
    // We need three arguments
    KiteReference<KiteObject> domain, type, protocol;
    
    if (typeid(*param) != typeid(KiteList) || ((KiteList&)(*param)).getArray().size() < 3) {
        return KiteReference<KiteObject>(new KiteTree("Not enough arguments"));
    }
    
    KiteList &list = (KiteList&)(*param);
    domain = coerceNumber(vm, list.getArray()[0]);
    type = coerceNumber(vm, list.getArray()[1]);
    protocol = coerceNumber(vm, list.getArray()[2]);
    
    KiteSocket *s = new KiteSocket(((KiteNumber&)(*domain)).intValue(), ((KiteNumber&)(*type)).intValue(),
                                   ((KiteNumber&)(*protocol)).intValue());
    
    if (s->getSocket() == -1) {
        // Throw an exception, error found
        KiteTree *t = new KiteTree(string("[socket] ") + strerror(errno));
        delete s;
        return KiteReference<KiteObject>(t);
    }
    
    // Otherwise, return the object we just created
    return KiteReference<KiteObject>(s);
}

KiteReference<KiteObject> KiteSocket_listen (VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> host, port;
    KiteSocket &s = (KiteSocket&)(*THIS_OBJECT());
    
    if (typeid(*param) == typeid(KiteList)) {
        KiteList &list = (KiteList&)(*param);
        host = coerceString(vm, list.getArray()[0]);
        port = coerceNumber(vm, list.getArray()[1]);
    } else {
        host = new KiteString("");
        port = coerceNumber(vm, param);
    }
    
    string h = ((KiteString&)(*host)).getString();
    int p = ((KiteNumber&)(*port)).intValue();
    
    struct hostent *server;
    struct sockaddr_in serv_addr;

    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_port = htons(p);
    serv_addr.sin_family = AF_INET;
    if (h == "") {
        serv_addr.sin_addr.s_addr = INADDR_ANY;
    } else {
        server = gethostbyname(h.c_str());
        if (server == NULL) {
            goto error;
        }
    
        bcopy((char *)server->h_addr, 
         (char *)&serv_addr.sin_addr.s_addr,
         server->h_length);
    }
    
    if (bind(s.getSocket(), (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
        goto error;
        
    if (listen(s.getSocket(), 5) < 0) {
        goto error;
    }
    
    return KiteReference<KiteObject>(new KiteBoolean(true));
error:
    return KiteReference<KiteObject>(new KiteTree(string("[socket] ") + strerror(errno)));
}

KiteReference<KiteObject> KiteSocket_accept (VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> newsocket, host, port;
    KiteSocket &s = (KiteSocket&)(*THIS_OBJECT());
    struct sockaddr_in cli_addr;
    
    KiteList *ret = new KiteList();
#ifdef WIN32
    int clilen = sizeof(cli_addr);
#else
    socklen_t clilen = sizeof(cli_addr);
#endif
    int client = accept(s.getSocket(), (struct sockaddr *) &cli_addr, &clilen);
    if (clilen < 0) {
        goto error;
    }
    
    port = new KiteNumber(ntohs(cli_addr.sin_port));
    host = new KiteString(inet_ntoa(cli_addr.sin_addr));
    newsocket = new KiteSocket(client);
    ret->getArray().push_back(newsocket);
    ret->getArray().push_back(host);
    ret->getArray().push_back(port);
    
    return KiteReference<KiteObject>(ret);
error:
    delete ret;
    return KiteReference<KiteObject>(new KiteTree(string("[socket] ") + strerror(errno)));
}

void KiteSocket::initialize() {
	syms["connect"] = KiteReference<KiteObject>(new KiteCode(KiteSocket_connect));
	syms["read"] = KiteReference<KiteObject>(new KiteCode(KiteSocket_read));
	syms["readline"] = KiteReference<KiteObject>(new KiteCode(KiteSocket_readline));
	syms["write"] = KiteReference<KiteObject>(new KiteCode(KiteSocket_write));
	syms["close"] = KiteReference<KiteObject>(new KiteCode(KiteSocket_close));
    
	syms["listen"] = KiteReference<KiteObject>(new KiteCode(KiteSocket_listen));
	syms["accept"] = KiteReference<KiteObject>(new KiteCode(KiteSocket_accept));
    
    // Constants
	syms["AF_UNIX"] = KiteReference<KiteObject>(new KiteNumber(AF_UNIX));
	syms["AF_INET"] = KiteReference<KiteObject>(new KiteNumber(AF_INET));
	syms["SOCK_STREAM"] = KiteReference<KiteObject>(new KiteNumber(SOCK_STREAM));
	syms["SOCK_DGRAM"] = KiteReference<KiteObject>(new KiteNumber(SOCK_DGRAM));
	syms["SOCK_RAW"] = KiteReference<KiteObject>(new KiteNumber(SOCK_RAW));
	syms["SOCK_SEQPACKET"] = KiteReference<KiteObject>(new KiteNumber(SOCK_SEQPACKET));
	syms["SOCK_RDM"] = KiteReference<KiteObject>(new KiteNumber(SOCK_RDM));
}
