#include <sstream>
#include <ctype.h>
#include "kite.h"
#include <FlexLexer.h>
#include "KiteVM.h"
#include "KiteNumber.h"
#include "KiteReference.h"
#include "KiteIdentifier.h"
#include "KiteCode.h"
#include "KiteBoolean.h"
#include "KiteTree.h"
#include "KiteString.h"
#include "KiteList.h"

using namespace KiteVM;

KiteReference<KiteObject> KiteString::classinfo;

KiteReference<KiteObject> KiteString_opplus(VM &vm, KiteReference<KiteObject> param) {
    KiteString &string1 = (KiteString&)(*coerceString(vm, THIS_OBJECT()));
    KiteReference<KiteObject> str2(coerceString(vm, param));
    KiteString &string2 = (KiteString&)(*str2);

    return KiteReference<KiteObject>(new KiteString(string1.getString() + string2.getString()));
}

KiteReference<KiteObject> KiteString_opmultiply(VM &vm, KiteReference<KiteObject> param) {
    KiteString &string = (KiteString&)(*coerceString(vm, THIS_OBJECT()));
    KiteReference<KiteObject> p2(coerceNumber(vm, param));
    KiteNumber &numtimes = (KiteNumber&)(*p2);
    std::string ret;

    for (int i = 0; i < numtimes.intValue(); i++) {
        ret += string.getString();
    }

    return KiteReference<KiteObject>(new KiteString(ret));
}

KiteReference<KiteObject> KiteString_print(VM &vm, KiteReference<KiteObject> param) {
    KiteString &str = (KiteString&)(*coerceString(vm, THIS_OBJECT()));

    cout << str.getString() << endl;
    return THIS_OBJECT();
}

KiteReference<KiteObject> KiteString_oplessthan(VM &vm, KiteReference<KiteObject> param) {
    KiteString &string1 = (KiteString&)(*coerceString(vm, THIS_OBJECT()));
    KiteReference<KiteObject> str2(coerceString(vm, param));
    KiteString &string2 = (KiteString&)(*str2);

    return KiteReference<KiteObject>(new KiteBoolean(string1 < string2));
}

KiteReference<KiteObject> KiteString_oplessequals(VM &vm, KiteReference<KiteObject> param) {
    KiteString &string1 = (KiteString&)(*coerceString(vm, THIS_OBJECT()));
    KiteReference<KiteObject> str2(coerceString(vm, param));
    KiteString &string2 = (KiteString&)(*str2);

    return KiteReference<KiteObject>(new KiteBoolean(string1 < string2 || string1 == string2));
}

KiteReference<KiteObject> KiteString_opequals(VM &vm, KiteReference<KiteObject> param) {
    KiteString &string1 = (KiteString&)(*coerceString(vm, THIS_OBJECT()));
    KiteReference<KiteObject> str2(coerceString(vm, param));
    KiteString &string2 = (KiteString&)(*str2);

    return KiteReference<KiteObject>(new KiteBoolean(string1 == string2));
}

KiteReference<KiteObject> KiteString_opnotequals(VM &vm, KiteReference<KiteObject> param) {
    KiteString &string1 = (KiteString&)(*coerceString(vm, THIS_OBJECT()));
    KiteReference<KiteObject> str2(coerceString(vm, param));
    KiteString &string2 = (KiteString&)(*str2);

    return KiteReference<KiteObject>(new KiteBoolean(!(string1 == string2)));
}

KiteReference<KiteObject> KiteString_opgreaterequals(VM &vm, KiteReference<KiteObject> param) {
    KiteString &string1 = (KiteString&)(*coerceString(vm, THIS_OBJECT()));
    KiteReference<KiteObject> str2(coerceString(vm, param));
    KiteString &string2 = (KiteString&)(*str2);

    return KiteReference<KiteObject>(new KiteBoolean(!(string1 < string2)));
}

KiteReference<KiteObject> KiteString_opgreaterthan(VM &vm, KiteReference<KiteObject> param) {
    KiteString &string1 = (KiteString&)(*coerceString(vm, THIS_OBJECT()));
    KiteReference<KiteObject> str2(coerceString(vm, param));
    KiteString &string2 = (KiteString&)(*str2);

    return KiteReference<KiteObject>(new KiteBoolean(!(string1 < string2) && !(string1 == string2)));
}

KiteReference<KiteObject> KiteString_number(VM &vm, KiteReference<KiteObject> param) {
    KiteString &string = (KiteString&)(*coerceString(vm, THIS_OBJECT()));

    if (string.getString().find(".") != std::string::npos) {
        return KiteReference<KiteObject>(new KiteNumber(atof(string.getString().c_str())));
    } else {
        return KiteReference<KiteObject>(new KiteNumber(atoi(string.getString().c_str())));
    }
}

KiteReference<KiteObject> KiteString_string(VM &vm, KiteReference<KiteObject> param) {
    return THIS_OBJECT();
}

KiteReference<KiteObject> KiteString_compile(VM &vm, KiteReference<KiteObject> param) {
    KiteString &string = (KiteString&)(*coerceString(vm, THIS_OBJECT()));
    stringstream str;
    
    str.str(string.getString());
    
    yyFlexLexer lexer(&str);
    ParserState state(&lexer);
    void *pParser;
    int hTokenId;
    
    state.file = "(string passed to compile)";
    try {
    pParser = ParseAlloc( malloc );
    while( (hTokenId = lexer.yylex()) != -1 ){
        if (hTokenId == NUMBER || hTokenId == STRING || hTokenId == IDENTIFIER) state.text = lexer.YYText();
        Parse(pParser, hTokenId, NULL, &state);
    }
    Parse(pParser, 0, NULL, &state);
    ParseFree(pParser, free );
    
    Instruction *i = state.root;
    if (i) {
        KiteReference<KiteObject> ret(new KiteCode(i));
        return ret;
    } else {
        return KiteReference<KiteObject>(new KiteBoolean(false));
    }
    } catch (KiteReference<KiteObject> &exc) {
        // Syntax error
        return exc;
    }
}

KiteReference<KiteObject> KiteString_eval(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> fxn = KiteString_compile(vm, param);
    
    // If there was a problem compiling, just return the exception.
    if (typeid(fxn) == typeid(KiteTree)) {
        return fxn;
    }
    
    // Otherwise, evaluate it.
    vm.syms.removeContext();
    KiteReference<KiteObject> ret(((KiteCode&)(*fxn)).call(vm, param));
    vm.syms.addContext();
    return ret;
}

KiteReference<KiteObject> KiteString_opderef(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> thisObj(THIS_OBJECT());
    KiteString &list = (KiteString&)(*thisObj);
    KiteReference<KiteObject> p2(coerceNumber(vm, param));
    KiteNumber &idx = (KiteNumber&)(*p2);
    char buf[2];
    stringstream str;
    
    if (idx.intValue() < 0 || idx.intValue() >= ((KiteString&)(*thisObj)).getString().size()) {
        str << idx.intValue();
        return KiteReference<KiteObject>(new KiteTree(string("Out of range: ") + str.str()));
    }

    buf[0] = list.getString()[idx.intValue() + 1];
    buf[1] = '\0';

    return KiteReference<KiteObject>(new KiteString(string(buf)));
}

KiteReference<KiteObject> KiteString_find(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> string(THIS_OBJECT());
    KiteReference<KiteObject> toFind(coerceString(vm, typeid(*param) == typeid(KiteList) ? ((KiteList&)(*param)).getArray()[0] : param));
    KiteReference<KiteObject> start;
    if (typeid(*param) == typeid(KiteList)) {
        start = coerceNumber(vm, ((KiteList&)(*param)).getArray()[1]);
    } else {
        start = KiteReference<KiteObject>(new KiteNumber(0));
    }

    KiteString &str = (KiteString&)(*toFind);
    int startPos = ((KiteNumber&)(*start)).intValue();
    if (startPos < 0 || startPos >= ((KiteString&)(*string)).getString().size()) {
        stringstream str;
        str << "Out of range: " << startPos;
        return KiteReference<KiteObject>(new KiteTree(str.str()));
    }
    int loc = ((KiteString&)(*string)).getString().find(str.getString(), startPos);
    
    return KiteReference<KiteObject>(new KiteNumber(loc));
}

KiteReference<KiteObject> KiteString_substring(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> string(THIS_OBJECT());
    KiteReference<KiteObject> begin(coerceNumber(vm, typeid(*param) == typeid(KiteList) ? ((KiteList&)(*param)).getArray()[0] : param));
    KiteReference<KiteObject> end;
    
    if (typeid(*param) == typeid(KiteList)) {
        end = coerceNumber(vm, ((KiteList&)(*param)).getArray()[1]);
    } else {
        end = KiteReference<KiteObject>(new KiteNumber(int(((KiteString&)(*string)).getString().size() - 1)));
    }

    int startPos = ((KiteNumber&)(*begin)).intValue();
    int endPos = ((KiteNumber&)(*end)).intValue();
    if (startPos < 0 || startPos >= ((KiteString&)(*string)).getString().size()) {
        stringstream str;
        str << "Out of range: " << startPos;
        return KiteReference<KiteObject>(new KiteTree(str.str()));
    }
    if (endPos < 0 || endPos >= ((KiteString&)(*string)).getString().size()) {
        stringstream str;
        str << "Out of range: " << endPos;
        return KiteReference<KiteObject>(new KiteTree(str.str()));
    }
    
    std::string res = ((KiteString&)(*string)).getString().substr(startPos, endPos);
    
    return KiteReference<KiteObject>(new KiteString(res));
}

KiteReference<KiteObject> KiteString_rfind(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> string(THIS_OBJECT());
    KiteReference<KiteObject> toFind(coerceString(vm, typeid(*param) == typeid(KiteList) ? ((KiteList&)(*param)).getArray()[0] : param));
    KiteReference<KiteObject> start;
    if (typeid(*param) == typeid(KiteList)) {
        start = coerceNumber(vm, ((KiteList&)(*param)).getArray()[1]);
    } else {
        start = KiteReference<KiteObject>(new KiteNumber(0));
    }

    KiteString &str = (KiteString&)(*toFind);
    int startPos = ((KiteNumber&)(*start)).intValue();
    int loc = ((KiteString&)(*string)).getString().rfind(str.getString(), startPos);
    
    return KiteReference<KiteObject>(new KiteNumber(loc));
}

KiteReference<KiteObject> KiteString_length(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> string(THIS_OBJECT());
    KiteString &str = (KiteString&)(*string);

    return KiteReference<KiteObject>(new KiteNumber((int)str.getString().size()));
}

KiteReference<KiteObject> KiteString_replace(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> string(THIS_OBJECT());
    
    if (typeid(*param) != typeid(KiteList) || ((KiteList&)*param).getArray().size() < 2) {
        return KiteReference<KiteObject>(new KiteTree("Not enough parameters (needed: 2)"));
    }
    
    KiteReference<KiteObject> toFind(coerceString(vm, ((KiteList&)(*param)).getArray()[0]));
    KiteReference<KiteObject> replacement(coerceString(vm, ((KiteList&)(*param)).getArray()[1]));
    KiteReference<KiteObject> start;
    
    if (((KiteList&)(*param)).getArray().size() == 3) {
        start = coerceNumber(vm, ((KiteList&)(*param)).getArray()[2]);
    } else {
        start = KiteReference<KiteObject>(new KiteNumber(0));
    }

    KiteString &str = (KiteString&)(*toFind);
    KiteString &replace = (KiteString&)(*replacement);
    int startPos = ((KiteNumber&)(*start)).intValue();
    if (startPos < 0 || startPos >= ((KiteString&)(*string)).getString().size()) {
        stringstream str;
        str << "Out of range: " << startPos;
        return KiteReference<KiteObject>(new KiteTree(str.str()));
    }
    
    std::string retstr = ((KiteString&)(*string)).getString();
    int loc;
    if ((loc = retstr.find(str.getString(), startPos)) != std::string::npos) {
        retstr.replace(loc, str.getString().size(), replace.getString());
    }
    
    return KiteReference<KiteObject>(new KiteString(retstr));
}

KiteReference<KiteObject> KiteString_lower(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> string(THIS_OBJECT());

    std::string curstr = ((KiteString&)(*string)).getString();
    std::string retstr;
    char *ptr = const_cast<char*>(curstr.c_str());
    while(*ptr) {
        retstr += tolower(*(ptr++));
    }
    return KiteReference<KiteObject>(new KiteString(retstr));
}

KiteReference<KiteObject> KiteString_upper(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> string(THIS_OBJECT());

    std::string curstr = ((KiteString&)(*string)).getString();
    std::string retstr;
    char *ptr = const_cast<char*>(curstr.c_str());
    while(*ptr) {
        retstr += toupper(*(ptr++));
    }
    return KiteReference<KiteObject>(new KiteString(retstr));
}

void KiteString::initialize() {
    syms["number"] = KiteReference<KiteObject>(new KiteCode(KiteString_number));
    syms["op+"] = KiteReference<KiteObject>(new KiteCode(KiteString_opplus));
    syms["op*"] = KiteReference<KiteObject>(new KiteCode(KiteString_opmultiply));
    syms["opderef"] = KiteReference<KiteObject>(new KiteCode(KiteString_opderef));

    syms["op<"] = KiteReference<KiteObject>(new KiteCode(KiteString_oplessthan));
    syms["op<="] = KiteReference<KiteObject>(new KiteCode(KiteString_oplessequals));
    syms["op=="] = KiteReference<KiteObject>(new KiteCode(KiteString_opequals));
    syms["op!="] = KiteReference<KiteObject>(new KiteCode(KiteString_opnotequals));
    syms["op>="] = KiteReference<KiteObject>(new KiteCode(KiteString_opgreaterequals));
    syms["op>"] = KiteReference<KiteObject>(new KiteCode(KiteString_opgreaterthan));

    //syms["xor"] = KiteReference<KiteObject>(new KiteCode(KiteString_opxor));

    syms["string"] = KiteReference<KiteObject>(new KiteCode(KiteString_string));
    syms["print"] = KiteReference<KiteObject>(new KiteCode(KiteString_print));
    
    syms["compile"] = KiteReference<KiteObject>(new KiteCode(KiteString_compile));
    syms["eval"] = KiteReference<KiteObject>(new KiteCode(KiteString_eval));
    
    syms["find"] = KiteReference<KiteObject>(new KiteCode(KiteString_find));
    syms["rfind"] = KiteReference<KiteObject>(new KiteCode(KiteString_rfind));
    syms["replace"] = KiteReference<KiteObject>(new KiteCode(KiteString_replace));
    syms["substring"] = KiteReference<KiteObject>(new KiteCode(KiteString_substring));
    syms["length"] = KiteReference<KiteObject>(new KiteCode(KiteString_length));
    syms["lower"] = KiteReference<KiteObject>(new KiteCode(KiteString_lower));
    syms["upper"] = KiteReference<KiteObject>(new KiteCode(KiteString_upper));
}
