/*
 *  KiteSystem.cpp
 *  kite
 *
 *  Created by Mooneer Salem on 12/30/06.
 *  Copyright 2006 __MyCompanyName__. All rights reserved.
 *
 */

#include <stdlib.h>
#include <string>
#include <sstream>
#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif
#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteSystem.h"
#include "KiteList.h"
#include "KiteCode.h"
#include "KiteString.h"
#include "KiteFile.h"
#include "KiteBoolean.h"
#include "KiteTree.h"
#include "KiteDynamicObject.h"

using namespace std;
using namespace KiteVM;

KiteReference<KiteObject> KiteSystem::classinfo;

KiteReference<KiteObject> KiteSystem_exit (VM &vm, KiteReference<KiteObject> param) {
    if (param.getObj() == NULL || (typeid(*param) == typeid(KiteList) && ((KiteList&)*param).getArray().size() == 0) ) exit(0);
    
    KiteReference<KiteObject> num(coerceNumber(vm, param));
    KiteNumber &s = (KiteNumber&)(*num);
    exit(s.intValue());
    return NULL;
}

KiteReference<KiteObject> KiteSystem_path (VM &vm, KiteReference<KiteObject> param) {
#ifdef WIN32
	const char* path_separator = ";";
#else
	const char* path_separator = ":";
#endif

    if (param.getObj() == NULL || (typeid(*param) == typeid(KiteList) && ((KiteList&)*param).getArray().size() == 0)) {
        // path retrieval mode
        char *path = strdup(getenv("PATH"));
        char *ap;
        KiteList *ret = new KiteList(false);
		ap = strtok(path, path_separator);
        while (ap != NULL) {
            ret->getArray().push_back(new KiteString(string(ap)));
			ap = strtok(NULL, path_separator);
        }
        free(path);
        return ret;
    }
    
    // path set mode
    stringstream str;
    KiteList *l;
    KiteReference<KiteObject> nl;
    if (typeid(*param) != typeid(KiteList)) {
        nl = new KiteList(param);
        l = (KiteList*)nl.getObj();
    } else {
        l = (KiteList*)param.getObj();
    }
    for (int i = 0; i < l->getArray().size(); i++) {
        str << ((KiteString&)*coerceString(vm, l->getArray().at(i))).getString() << ":";
    }
    string res = str.str().substr(0, str.str().size() - 1);
#ifdef WIN32
	WCHAR *wszName = new WCHAR[5]; // PATH + null
	WCHAR *wszValue = new WCHAR[res.size() + 1];
	MultiByteToWideChar(CP_ACP, 0, "PATH", 5, wszName, sizeof(wszName)/sizeof(wszName[0]));
	MultiByteToWideChar(CP_ACP, 0, res.c_str(), res.size()+1, wszValue, sizeof(wszValue)/sizeof(wszValue[0]));
	SetEnvironmentVariable(wszName, wszValue);
	delete[] wszName;
	delete[] wszValue;
#else
    setenv("PATH", res.c_str(), 1);
#endif

    return KiteReference<KiteObject>(new KiteBoolean(true));
}

KiteReference<KiteObject> KiteSystem_getenv (VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> var(coerceString(vm, param));
    
    return KiteReference<KiteObject>(new KiteString(getenv(((KiteString&)*var).getString().c_str())));
}

KiteReference<KiteObject> KiteSystem_setenv (VM &vm, KiteReference<KiteObject> param) {
    if (param.getObj() == NULL || typeid(*param) != typeid(KiteList)) {
notenough:
        return KiteReference<KiteObject>(new KiteTree("Not enough parameters"));
    }
    
    KiteList &l = (KiteList&)*param;
    if (l.getArray().size() < 2) goto notenough;
    KiteReference<KiteObject> name(coerceString(vm, l.getArray()[0]));
    KiteReference<KiteObject> val(coerceString(vm, l.getArray()[1]));
    string n = ((KiteString&)*name).getString();
    string v = ((KiteString&)*val).getString();

#ifdef WIN32
	WCHAR *wszName = new WCHAR[n.size() + 1];
	WCHAR *wszValue = new WCHAR[v.size() + 1];
	MultiByteToWideChar(CP_ACP, 0, n.c_str(), n.size()+1, wszName, sizeof(wszName)/sizeof(wszName[0]));
	MultiByteToWideChar(CP_ACP, 0, v.c_str(), v.size()+1, wszValue, sizeof(wszValue)/sizeof(wszValue[0]));
	SetEnvironmentVariable(wszName, wszValue);
	delete[] wszName;
	delete[] wszValue;
#else
    setenv(n.c_str(), v.c_str(), 1);
#endif
    return KiteReference<KiteObject>(new KiteBoolean(true));
}

KiteReference<KiteObject> KiteSystem_load (VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> name(coerceString(vm, param));
    
    return KiteReference<KiteObject>(new KiteDynamicObject(vm, ((KiteString&)*name).getString().c_str()));
}

void KiteSystem::initialize() {
	syms["exit"] = KiteReference<KiteObject>(new KiteCode(KiteSystem_exit));
    syms["path"] = KiteReference<KiteObject>(new KiteCode(KiteSystem_path));
    syms["getenv"] = KiteReference<KiteObject>(new KiteCode(KiteSystem_getenv));
    syms["setenv"] = KiteReference<KiteObject>(new KiteCode(KiteSystem_setenv));
	syms["load"] = KiteReference<KiteObject>(new KiteCode(KiteSystem_load));

    // Constants
    syms["in"] = KiteReference<KiteObject>(new KiteFile(&std::cin));
    syms["out"] = KiteReference<KiteObject>(new KiteFile(&std::cout));
	syms["err"] = KiteReference<KiteObject>(new KiteFile(&std::cerr));
}
