#ifndef KITETREE_H
#define KITETREE_H

#include <string>
#include "KiteIdentifier.h"
#include "KiteString.h"
#include "KiteNumber.h"

using namespace KiteVM;

class KiteTree : public KiteObject {
public:
	KiteTree()
		: KiteObject(false) {
        inherits = KiteObject::getClassinfo();
        this->type = "tree";
		initialize();
	}

	KiteTree(std::string str) : KiteObject(false),
		instance(true), msg(str) {
        inherits = KiteTree::getClassinfo();
        this->type = "tree";
    }

	std::string getMessage() { return msg; }

    void setBacktrace(vector<FuncCall> bt) { backtrace = bt; }
    vector<FuncCall> &getBacktrace() { return backtrace; }
    
    virtual KiteReference<KiteObject> getClassinfo() {
        if (classinfo.getObj() == NULL) {
            classinfo = new KiteTree();
        }
        return classinfo;
    }
private:
	std::string msg;
	bool instance;
	void initialize();
    vector<FuncCall> backtrace;
    static KiteReference<KiteObject> classinfo;
};

class KiteTrace : public KiteObject {
public:
    KiteTrace() : KiteObject(false) {
        inherits = KiteObject::getClassinfo();
        this->type = "trace";
        initialize();
    }
    
    KiteTrace(FuncCall &c) : KiteObject(false), fc(c) {
        inherits = KiteTrace::getClassinfo();
        this->type = "trace";
        
        syms["file"] = KiteReference<KiteObject>(new KiteString(fc.file));
        syms["name"] = KiteReference<KiteObject>(new KiteString(fc.name));
        syms["line"] = KiteReference<KiteObject>(new KiteNumber(fc.line));
    }
    
    virtual KiteReference<KiteObject> getClassinfo() {
        if (classinfo.getObj() == NULL) {
            classinfo = new KiteTrace();
        }
        return classinfo;
    }
    
    FuncCall &getFc() { return fc; }
private:
    FuncCall fc;
    bool instance;
    void initialize();
    static KiteReference<KiteObject> classinfo;
};

#endif // KITETREE_H

