%{
#include <iostream>
#include "kite.h"
#include "ParserState.h"

extern ParserState *state;
%}

NUMBER      [0-9]+(\.[0-9]+([Ee]-?[0-9]+)?)?
IDENTIFIER  [A-Za-z_][^, \r\n\t;\(\)\|\.\[\]]*
DSTRING     \"([^\"]*|\\([\\rnt"]|x?[0-9]+))\"
SSTRING     '([^\']*|\\([\\rnt']|x?[0-9]+))'
COMMENT     #[^\r\n]*
WS          [ \t]+

%option c++
%option yylineno
%option noyywrap

%%

"and"       return AND;
"or"        return OR;
"not"       return NOT;
"mod"       return MOD;
"load"      return LOAD;
"var"       return VAR;
"class"     return CLASS;
"true"      return TRUE;
"false"     return FALSE;

{NUMBER}    return NUMBER;
{IDENTIFIER} return IDENTIFIER;
{DSTRING}|{SSTRING} return STRING;
{COMMENT}|{WS} { }
\r|\n|\r\n  return 256;
"+"         return PLUS;
"-"         return MINUS;
"*"         return STAR;
"/"         return DIVIDE;
";"         return SEMICOLON;
"."         return DEREFERENCE;
"|"         return FUNCCALL;
"<<"        return LEFTSHIFT;
">>"        return RIGHTSHIFT;
"<="        return LE;
"<"         return LESSTHAN;
">="        return GE;
">"         return GREATERTHAN;
"=="        return EQUALITY;
"!="        return NOTEQUALS;
"{"         {
                if (state) state->bracelevel++;
                return LCURLY;
            }
"}"         {
                if (state) state->bracelevel--;
                return RCURLY;
            }
"="         return EQUALS;
"("         {
                if (state) state->bracelevel++;
                return LPAREN;
            }
")"         {
                if (state) state->bracelevel--;
                return RPAREN;
            }
","         return COMMA;
":"         return COLON;
"["         {
                if (state) state->bracelevel++;
                return LBRACKET;
            }
"]"         {
                if (state) state->bracelevel--;
                return RBRACKET;
            }
.           return -65535;
%%
