#include <iostream>
#include <fstream>
#include <FlexLexer.h>
#include "kite.h"
#include "KiteVM.h"
#include "KiteIdentifier.h"
#include "KiteObject.h"
#include "KiteTree.h"
#include "config.h"

using namespace std;
using namespace KiteVM;

#ifdef ALTERNATE_ALLOCATOR
const int ALTERNATE_THRESHOLD = 16;

struct entry {
    char data[ALTERNATE_THRESHOLD];
    unsigned int size;
    unsigned int head;
    unsigned int flags;
};

static struct entry **head = NULL;
static int cur, curhead, numheads = 0;

const int PAGE_SIZE = 32767;
const int NUM_PER_PAGE = PAGE_SIZE / sizeof(struct entry);

void *operator new(size_t size) {
    if (size > ALTERNATE_THRESHOLD) {
        return malloc(size);
    }

    if (!head) {
        head = (struct entry**)malloc(sizeof(void*));
        head[0] = (struct entry*)malloc(PAGE_SIZE);
        head[0]->flags = 0;
        head[0]->size = ALTERNATE_THRESHOLD;
        head[0]->head = 0;
        curhead = 0;
        cur = 0;
        numheads++;
    }

top:
    for ( ; cur < NUM_PER_PAGE; cur++) {
        if (head[curhead][cur].flags) continue;
        else break;
    }

    if (cur == NUM_PER_PAGE) {
        if (curhead < (numheads - 1)) {
            curhead++;
            cur = 0;
            goto top;
        }
        numheads++;
        head = (struct entry**)realloc(head, sizeof(struct entry**)*numheads);
        curhead = numheads - 1;
        head[curhead] = (struct entry*)malloc(PAGE_SIZE);
        head[curhead]->flags = 0;
        head[curhead]->size = ALTERNATE_THRESHOLD;
        head[curhead]->head = curhead;
        cur = 0;
    }

    assert(curhead < numheads && cur < NUM_PER_PAGE);
    head[curhead][cur].size = ALTERNATE_THRESHOLD;
    head[curhead][cur].flags = 1;
    head[curhead][cur].head = curhead;

    void *ret = &head[curhead][cur].data;
    cur++;
    return ret;
}

void operator delete(void *ptr, size_t size) {
    if (size > ALTERNATE_THRESHOLD) {
    	free(ptr);
        return;
    }

    struct entry *e = (struct entry*)ptr;
    e->flags = 0;
    /*curhead = e->head;
    cur = (e - head[e->head]) / sizeof(struct entry);*/
}

void operator delete(void *ptr) {
    struct entry *e = (struct entry*)ptr;
    if (e->size <= ALTERNATE_THRESHOLD) operator delete(ptr, e->size);
    else free(ptr);
}
#endif

ParserState *state = NULL;

int main(int argc, char **argv) {
    yyFlexLexer *lexer = NULL;
    void *pParser;
    int hTokenId;
    bool interactive = false;
    
    // Use standard input if nothing passed in
    ifstream input;
    state = new ParserState();
    if (argc == 2) {
        input.open(argv[1], ios::in);
        lexer = new yyFlexLexer(&input);
        state->file = argv[1];
    } else {
        state->file = "(standard input)";
        interactive = true;
        lexer = new yyFlexLexer(&cin);
    }
    
    if (interactive) {
        // Copyright lines
        cout << "Kite version " << PACKAGE_VERSION << endl;
        cout << "Copyright (C) 2006-2007 Mooneer Salem" << endl;
    }
    
    KiteVM::VM vm;
    do {
        pParser = ParseAlloc(malloc);
        try {
standard_input:
            if (interactive) {
                // Get a single line from cin.
                cout << "[" << state->bracelevel << "]> ";
            }

            state->lexer = lexer; 
            while( (hTokenId = lexer->yylex()) != 0 ){
                if (hTokenId == NUMBER || hTokenId == STRING || hTokenId == IDENTIFIER) state->text = lexer->YYText();
                if (hTokenId == 256 && interactive && state->bracelevel) goto standard_input;
                else if (hTokenId == 256 && !interactive) continue;
                else if (hTokenId == 256 && interactive) break;
                Parse(pParser, hTokenId, NULL, state);
            }
            
            Parse(pParser, 0, NULL, state);
            
            KiteVM::Instruction *i = state->root;
            if (i) {
                KiteReference<KiteObject> ret = vm.run(i);
                
                if (typeid(*ret) == typeid(KiteTree)) {
                    throw ret;
                }
                
                if (interactive) {
                    vm.syms.addContext();
                    vm.syms.addVar("this", ret);
                    KiteReference<KiteObject> printMethod = (*ret).dereference(vm, "print");
                    (*printMethod).call(vm, ret);
                    vm.syms.removeContext();
                }

                i->deallocate();
            }
        } catch(KiteReference<KiteObject> &e) {
            KiteTree &exc = (KiteTree&)(*e);
            cerr << "Exception in code: " << exc.getMessage() << endl;
            vector<FuncCall> &bt = exc.getBacktrace();
            for (int j = bt.size() - 1; j >= 0; j--) {
                cerr << "  at " << bt.at(j).file << "(" << bt.at(j).name << ") line " << bt.at(j).line << endl;
            }
        }
        ParseFree(pParser, free);
    } while (interactive && !cin.eof());
    
    delete lexer;
    delete state;
    
    return 0;
}
